#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <dmedia/audio.h>
#include <errno.h>
#include "audio.h"

typedef signed char audio_samp_t;

#define DEFAULT_VOL 100

static ALconfig config;
static ALport   port = NULL;
static stamp_t  start_samples;
#define nchannel 2

typedef struct _AudioChannel
{
    audio_samp_t* samp;
    int nsamp;			/* number of samples */
    int beg;			/* start samples */
    int off;
    int vol;			/* percent */
} AudioChannel;

static AudioChannel audio_channel[MAX_CHANNEL];

void audio_open(void)
{
    ALpv pvs[1];
    int r;
    int i;

    if(port != NULL)
	return;

    config = alNewConfig();

    /* Linear PCM two's complement integer */
    alSetSampFmt(config, AL_SAMPFMT_TWOSCOMP);

    if(alSetWidth(config, AL_SAMPLE_8) < 0)
    {
	fprintf(stderr, "Couldn't set width:%s\n",
		alGetErrorString(oserror()));
	exit(1);
    }

    if(alSetChannels(config, nchannel == 2 ? AL_STEREO : AL_MONO) < 0)
    {
	fprintf(stderr, "Couldn't set channel:%s\n",
		alGetErrorString(oserror()));
	exit(1);
    }

    /* Now attempt to open an audio output port using this config */
    port = alOpenPort("Audio", "w", config);
    if(!port)
    {
	fprintf(stderr, "Couldn't open an audio port:%s\n",
		alGetErrorString(oserror()));
	exit(1);
    }

    if((r = alGetResource(port)) < 0)
    {
	fprintf(stderr, "Couldn't get a device resource: %s\n",
		alGetErrorString(oserror()));
	exit(1);
    }

    pvs[0].param = AL_RATE;
    pvs[0].value.ll = alIntToFixed(AL_RATE_8000);
    if(alSetParams(r, pvs, 1) < 0)
    {
	fprintf(stderr, "Couldn't set a device resource: %s\n",
		alGetErrorString(oserror()));
	exit(1);
    }

    memset(audio_channel, 0, sizeof(audio_channel));

    {
	stamp_t dmy;
	alGetFrameTime(port, &start_samples, &dmy);    
    }

    for(i = 0; i < MAX_CHANNEL; i++)
	if(audio_channel[i].vol == 0)
	    audio_channel[i].vol = DEFAULT_VOL;
}

void audio_close(void)
{
    if(port != NULL)
    {
	alClosePort(port);
	port = NULL;
    }
}

static long audriv_play_samples(void)
/* ߱Υץ֤֤ޤ
 */
{
    stamp_t s, t;
    long ret;
    struct
    {
	unsigned long l1;
	unsigned long l2;
    } l;

    alGetFrameTime(port, &s, &t);
    s -= start_samples;

    memcpy(&l, &s, 8);
    ret = (long)s;

    return ret;
}

void play_stop(int cid)
{
    stamp_t s, dmy;
    long pos;
    int i;

    if(port == NULL)
	return;

    pos = audriv_play_samples() * nchannel;

    alDiscardFrames(port, alGetFilled(port));
    alGetFrameTime(port, &s, &dmy);
    start_samples = s;

    if(cid == -1)
    {
	for(i = 0; i < MAX_CHANNEL; i++)
	{
	    AudioChannel* c;

	    c = audio_channel + i;
	    if(c->samp)
	    {
		long cpos;
		cpos = c->off + pos;
		if(pos < c->nsamp)
		{
		    c->beg = cpos;
		    c->off = cpos;
		}
		else
		    c->samp = NULL;
	    }
	}
    }
    else
    {
	AudioChannel* c;

	c = audio_channel + cid;
	c->samp = NULL;
	c->beg = 0;
	c->off = 0;
	play_loop();
    }
}

#define BUFFSIZE 4196
void play_loop(void)
{
    short outbuff[BUFFSIZE];
    audio_samp_t* op;
    int i, outlen, fillable;

    if(port == NULL)
	return;

    memset(outbuff, 0, sizeof(outbuff));
    fillable = alGetFillable(port) * nchannel;
    if(fillable > BUFFSIZE)
	fillable = BUFFSIZE;
    outlen = 0;

    for(i = 0; i < MAX_CHANNEL; i++)
    {
	AudioChannel* c;
	c = audio_channel + i;

	if(c->samp && c->beg < c->nsamp)
	{
	    int j, n, v;
	    audio_samp_t* s;

	    n = c->nsamp - c->beg;
	    v = c->vol;
	    if(n > fillable)
		n = fillable;
	    s = c->samp + c->beg;

	    if(v == DEFAULT_VOL)
	    {
		for(j = 0; j < n; j++)
		    outbuff[j] += s[j];
	    }
	    else
	    {
		double vr;
		vr = (double)v / (double)DEFAULT_VOL;
		for(j = 0; j < n; j++)
		    outbuff[j] += s[j] * vr;
	    }
	    c->beg += n;
	    if(outlen < n)
		outlen = n;
	}
    }

    op = (audio_samp_t *)outbuff;
    for(i = 0; i < outlen; i++)
    {
	if(outbuff[i] < -128)
	    op[i] = -128;
	else if(outbuff[i] > 127)
	    op[i] = 127;
	else
	    op[i] = outbuff[i];
    }
    alWriteFrames(port, op, outlen / nchannel);
}

void play(int cid, char* samp, int nsamp)
{
    AudioChannel* c;

    if(port == NULL)
	return;

    c = audio_channel + cid;
    play_stop(-1);
    c->samp = (audio_samp_t *)samp;
    c->nsamp = nsamp;
    c->beg = 0;
    c->off = 0;
    play_loop();
}

int audio_is_active(int cid)
{
    AudioChannel* c;
    long pos;

    if(port == NULL)
	return 0;

    c = audio_channel + cid;
    if(c->samp == NULL)
	return NULL;
    pos = audriv_play_samples() * nchannel;
    return c->off + pos < c->nsamp;
}


void audio_volume(int cid, int vol)
{
    if(cid == -1)
    {
	int i;
	for(i = 0; i < MAX_CHANNEL; i++)
	    audio_channel[i].vol = vol;
    }
    else
	audio_channel[cid].vol = vol;
}

int get_audio_volume(int cid)
{
    return audio_channel[cid].vol;
}
